import sys
import argparse
import logging
import time

def pase_info(seq):
	info = {'SVLEN': 0, 'END': 0, "SVTYPE": '', "RE": 0, "CHR2": ''}
	for i in seq.split(';'):
		if i.split('=')[0] in ["SVLEN", "END", "RE"]:
			try:
				info[i.split('=')[0]] = abs(int(float(i.split('=')[1])))
			except:
				pass
		if i.split('=')[0] in ["CHR2"]:
			info[i.split('=')[0]] = i.split('=')[1]
		if i.split('=')[0] in ["SVTYPE"]:
			info[i.split('=')[0]] = i.split('=')[1][0:3]

	return info

def phase_GT(seq):
	i = seq.split(':')
	if i[0] in ['0/1', '1/0']:
		return 'het'
	elif i[0] == '1/1':
		return 'hom'
	else:
		return 'unknown'


def load_callset(path, svtype_list):
	callset = dict()
	abtype = dict()
	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		if seq[0][0] == '#':
			continue

		chr = seq[0]
		pos = int(seq[1])
		info = pase_info(seq[7])

		if len(svtype_list) == 3 and info['SVTYPE'] == "DUP":
			info['SVTYPE'] = "INS"

		if info['SVTYPE'] in svtype_list:
			if info['SVTYPE'] == "BND":
				if seq[4][0] == ']':
					form = ']]N'
					chr2 = seq[4].split(':')[0][1:]
					pos2 = int(seq[4].split(':')[1][:-2])
				elif seq[4][0] == '[':
					form = '[[N'
					chr2 = seq[4].split(':')[0][1:]
					pos2 = int(seq[4].split(':')[1][:-2])
				else:
					if seq[4][1] == ']':
						form = 'N]]'
						chr2 = seq[4].split(':')[0][2:]
						pos2 = int(seq[4].split(':')[1][:-1])
					else:
						form = 'N[['
						chr2 = seq[4].split(':')[0][2:]
						pos2 = int(seq[4].split(':')[1][:-1])
				if info['SVTYPE'] not in callset:
					callset[info['SVTYPE']] = list()
				if info['END'] == 0:
					info['CHR2'] = chr2
					info['END'] = pos2
				try:
					if int(chr) <= int(info['CHR2']):
						if form == 'N[[':
							form = ']]N'
						if form == ']]N':
							form = 'N[['
						callset[info['SVTYPE']].append([chr, pos, info['CHR2'], info['END'], form, phase_GT(seq[9]), 0])
					else:
						callset[info['SVTYPE']].append([info['CHR2'], info['END'], chr, pos, form, phase_GT(seq[9]), 0])
				except:
					callset[info['SVTYPE']].append([chr, pos, info['CHR2'], info['END'], form, phase_GT(seq[9]), 0])

			else:
				if info['SVTYPE'] not in callset:
					callset[info['SVTYPE']] = list()
				if info['SVLEN'] == 0:
					info['SVLEN'] = info['END'] - pos + 1
				callset[info['SVTYPE']].append([chr, pos, info['END'], info['SVLEN'], phase_GT(seq[9]), 0])
		else:
			if info['SVTYPE'] not in abtype:
				abtype[info['SVTYPE']] = 0
			abtype[info['SVTYPE']] += 1

	file.close()
	return callset, abtype

def eval(call, ans, bias, offect, opt, genotype):
	for svtype in call:
		if svtype not in ans:
			# continue
			if svtype == 'INS':
				for i in call[svtype]:
					for key in ans:
						for j in ans[key]:
							if i[0] == j[0]:
								if abs(i[1] - j[1]) <= offect and float(min(i[3],j[3])/max(i[3],j[3])) >= bias:
									i[-1] = 1
									j[3+opt] = 1
									if i[4] == genotype[j[0]]:
										i[-1] = 2
										j[3+opt] = 2
		else:
			for i in call[svtype]:
				for j in ans[svtype]:
					if i[0] != j[0]:
						continue
					else:
						if svtype in ['INS']:
							if abs(i[1] - j[1]) <= offect and float(min(i[3],j[2])/max(i[3],j[2])) >= bias:
								j[2+opt] = 1
								i[-1] = 1
								if i[4] == genotype[j[0]]:
									j[2+opt] = 2
									i[-1] = 2
						elif svtype == 'BND':
							if i[2] != j[2]:
								continue
							else:
								# if i[4] == j[4]:
								if 1:
									if abs(i[1]-j[1]) <= offect and abs(i[3]-j[3]) <= offect:
										i[-1] = 1
										j[4+opt] = 1
										if i[5] == genotype[j[0]] or i[5] == genotype[j[2]]:
											i[-1] = 2
											j[4+opt] = 2
						else:
							if max(i[1]-offect, j[1]) <= min(i[2]+offect, j[2]) and float(min(i[3],j[3])/max(i[3],j[3])) >= bias:
								j[3+opt] = 1
								i[-1] = 1
								if i[4] == genotype[j[0]]:
									j[3+opt] = 2
									i[-1] = 2
						# genotype


def statistics(call, ans, opt, res):
	for svtype in call:
		tp = 0
		total = 0
		for ele in call[svtype]:
			total += 1
			if ele[-1] >= res:
				tp += 1
			# if ele[-1] < res and ele[-1] == 1:
			# 	print(ele)
		logging.info('TP-%d of %s:\t%d\t%d'%(res, svtype, tp, total))

	for svtype in ans:
		fn = 0
		total = 0
		for ele in ans[svtype]:
			total += 1
			if svtype == 'INS':
				if ele[2+opt] >= res:
					fn += 1
			elif svtype == 'BND':
				if ele[4+opt] >= res:
					fn += 1
			else:
				if ele[3+opt] >= res:
					fn += 1
		logging.info('TN-%d of %s:\t%d\t%d'%(res, svtype, fn, total))

typetrans = {'insertion':'INS', 
			'deletion':'DEL', 
			'inversion':'INV',
			'tandem duplication':'DUP',
			'reciprocal translocation':'BND'
			}

def load_ans(path):
	ansbed = dict()
	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		chr = seq[0]
		svtype = typetrans[seq[3]]
		start = int(seq[1])
		end = int(seq[2])

		if svtype not in ansbed:
			ansbed[svtype] = list()

		if svtype in ['INS']:
			ansbed[svtype].append([chr, start, len(seq[4]), 0, 0, 0, 0])
		elif svtype in ['BND']:
			chr2 = seq[4].split(':')[1]
			start2 = int(seq[4].split(':')[2])
			strand1 = seq[4].split(':')[3]
			strand2 = seq[4].split(':')[4]

			if strand1[0] == 'f':
				if strand2[0] == 'f':
					ansbed[svtype].append([chr, start, chr2, start2, "N[[", 0, 0, 0, 0])
					# ansbed[svtype].append([chr, start, chr2, start2, "]]N", 0, 0, 0, 0])
					# ansbed[svtype].append([chr, end, chr2, start2+end-start, "]]N", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2+end-start, "N[[", 0, 0, 0, 0])
				else:
					ansbed[svtype].append([chr, start, chr2, start2, "N[[", 0, 0, 0, 0])
					ansbed[svtype].append([chr, start, chr2, start2+end-start, "[[N", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2, "N]]", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2+end-start, "]]N", 0, 0, 0, 0])
			else:
				if strand2[0] == 'f':
					ansbed[svtype].append([chr, start, chr2, start2+end-start, "N]]", 0, 0, 0, 0])
					ansbed[svtype].append([chr, start, chr2, start2, "]]N", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2, "[[N", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2+end-start, "N[[", 0, 0, 0, 0])
				else:
					ansbed[svtype].append([chr, start, chr2, start2+end-start, "N]]", 0, 0, 0, 0])
					ansbed[svtype].append([chr, end, chr2, start2, "N]]", 0, 0, 0, 0])
					# ansbed[svtype].append([chr, end, chr2, start2, "[[N", 0, 0, 0, 0])
					# ansbed[svtype].append([chr, start, chr2, start2+end-start, "[[N", 0, 0, 0, 0])
		else:
			ansbed[svtype].append([chr, start, end, end-start+1, 0, 0, 0, 0])

	file.close()
	return ansbed

def load_gt(path):
	GT = dict()
	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		chr = seq[0]
		if chr not in GT:
			GT[chr] = ''
		if float(seq[-1]) > 80.0:
			GT[chr] = 'hom'
		elif 80.0 >= float(seq[-1]) > 20.0:
			GT[chr] = 'het'
		else:
			GT[chr] = 'None'
	return GT

def main_ctrl(args):

	# load ground truth set
	ans = load_ans(args.ans)
	genotype = load_gt(args.gt)

	if args.choice == "BND":
		cuteSV, abcuteSV = load_callset(args.cuteSV, ["BND"])
		logging.info("The number of calls within abnormal SV type in cuteSV:")
		for key in abcuteSV:
			logging.info("<cuteSV-%s>\t%d."%(key, abcuteSV[key]))
		logging.info("Evaluation on cuteSV callsets...")
		eval(cuteSV, ans, args.bias, args.offect, 1, genotype)
		statistics(cuteSV, ans, 1, 1)
		statistics(cuteSV, ans, 1, 2)

		sniffles, absniffles = load_callset(args.sniffles, ["BND"])
		logging.info("The number of calls within abnormal SV type in Sniffles:")
		for key in absniffles:
			logging.info("<Sniffles-%s>\t%d."%(key, absniffles[key]))
		logging.info("Evaluation on Sniffles callsets...")
		eval(sniffles, ans, args.bias, args.offect, 2, genotype)
		statistics(sniffles, ans, 2, 1)
		statistics(sniffles, ans, 2, 2)

		pbsv, abpbsv = load_callset(args.pbsv, ["BND"])
		logging.info("The number of calls within abnormal SV type in PBSV:")
		for key in abpbsv:
			logging.info("<PBSV-%s>\t%d."%(key, abpbsv[key]))
		logging.info("Evaluation on PBSV callsets...")
		eval(pbsv, ans, args.bias, args.offect, 3, genotype)
		statistics(pbsv, ans, 3, 1)
		statistics(pbsv, ans, 3, 2)

		svim, absvim = load_callset(args.svim, ["BND"])
		logging.info("The number of calls within abnormal SV type in SVIM:")
		for key in absvim:
			logging.info("<SVIM-%s>\t%d."%(key, absvim[key]))
		logging.info("Evaluation on SVIM callsets...")
		eval(svim, ans, args.bias, args.offect, 4, genotype)
		statistics(svim, ans, 4, 1)
		statistics(svim, ans, 4, 2)

	if args.choice == "DUP":
		cuteSV, abcuteSV = load_callset(args.cuteSV, ["INS", "DUP"])
		logging.info("The number of calls within abnormal SV type in cuteSV:")
		for key in abcuteSV:
			logging.info("<cuteSV-%s>\t%d."%(key, abcuteSV[key]))
		logging.info("Evaluation on cuteSV callsets...")
		eval(cuteSV, ans, args.bias, args.offect, 1, genotype)
		statistics(cuteSV, ans, 1, 1)
		statistics(cuteSV, ans, 1, 2)

		sniffles, absniffles = load_callset(args.sniffles, ["INS", "DUP"])
		logging.info("The number of calls within abnormal SV type in Sniffles:")
		for key in absniffles:
			logging.info("<Sniffles-%s>\t%d."%(key, absniffles[key]))
		logging.info("Evaluation on Sniffles callsets...")
		eval(sniffles, ans, args.bias, args.offect, 2, genotype)
		statistics(sniffles, ans, 2, 1)
		statistics(sniffles, ans, 2, 2)

		pbsv, abpbsv = load_callset(args.pbsv, ["INS", "DUP"])
		logging.info("The number of calls within abnormal SV type in PBSV:")
		for key in abpbsv:
			logging.info("<PBSV-%s>\t%d."%(key, abpbsv[key]))
		logging.info("Evaluation on PBSV callsets...")
		eval(pbsv, ans, args.bias, args.offect, 3, genotype)
		statistics(pbsv, ans, 3, 1)
		statistics(pbsv, ans, 3, 2)

		svim, absvim = load_callset(args.svim, ["INS", "DUP"])
		logging.info("The number of calls within abnormal SV type in SVIM:")
		for key in absvim:
			logging.info("<SVIM-%s>\t%d."%(key, absvim[key]))
		logging.info("Evaluation on SVIM callsets...")
		eval(svim, ans, args.bias, args.offect, 4, genotype)
		statistics(svim, ans, 4, 1)
		statistics(svim, ans, 4, 2)

	if args.choice == "IID":
		
		cuteSV, abcuteSV = load_callset(args.cuteSV, ["INS", "INV", "DEL"])
		logging.info("The number of calls within abnormal SV type in cuteSV:")
		for key in abcuteSV:
			logging.info("<cuteSV-%s>\t%d."%(key, abcuteSV[key]))
		logging.info("Evaluation on cuteSV callsets...")
		eval(cuteSV, ans, args.bias, args.offect, 1, genotype)
		statistics(cuteSV, ans, 1, 1)
		statistics(cuteSV, ans, 1, 2)

		sniffles, absniffles = load_callset(args.sniffles, ["INS", "INV", "DEL"])
		logging.info("The number of calls within abnormal SV type in Sniffles:")
		for key in absniffles:
			logging.info("<Sniffles-%s>\t%d."%(key, absniffles[key]))
		logging.info("Evaluation on Sniffles callsets...")
		eval(sniffles, ans, args.bias, args.offect, 2, genotype)
		statistics(sniffles, ans, 2, 1)
		statistics(sniffles, ans, 2, 2)

		pbsv, abpbsv = load_callset(args.pbsv, ["INS", "INV", "DEL"])
		logging.info("The number of calls within abnormal SV type in PBSV:")
		for key in abpbsv:
			logging.info("<PBSV-%s>\t%d."%(key, abpbsv[key]))
		logging.info("Evaluation on PBSV callsets...")
		eval(pbsv, ans, args.bias, args.offect, 3, genotype)
		statistics(pbsv, ans, 3, 1)
		statistics(pbsv, ans, 3, 2)

		svim, absvim = load_callset(args.svim, ["INS", "INV", "DEL"])
		logging.info("The number of calls within abnormal SV type in SVIM:")
		for key in absvim:
			logging.info("<SVIM-%s>\t%d."%(key, absvim[key]))
		logging.info("Evaluation on SVIM callsets...")
		eval(svim, ans, args.bias, args.offect, 4, genotype)
		statistics(svim, ans, 4, 1)
		statistics(svim, ans, 4, 2)

		# ***********optional output****************
		# '''
		# for key in ans:
		# 	for i in ans[key]:
		# 		if key == 'INS':
		# 			if i[3] == 0:
		# 				print('%s\t%d\t%s\t%d'%(i[0], i[1], key, i[2]))
		# 		elif key == 'DEL':
		# 			if i[4] == 0:
		# 				print('%s\t%d\t%s\t%d'%(i[0], i[1], key, i[3]))
		# 		else:
		# 			if i[4] == 0:
		# 				print(key, i)
		# '''
		# for key in cuteSV:
		# 	for i in cuteSV[key]:
		# 		if key == 'INS':
		# 			if i[4] == 0:
		# 				print('%s\t%d\t%s\t%d'%(i[0], i[1], key, i[3]))
		# 		else:
		# 			if i[4] == 0:
		# 				print('%s\t%d\t%s\t%d'%(i[0], i[1], key, i[3]))


def main(argv):
	args = parseArgs(argv)
	setupLogging(False)
	# print args
	starttime = time.time()
	main_ctrl(args)
	logging.info("Finished in %0.2f seconds."%(time.time() - starttime))

USAGE="""\
	Evaluate SV callset generated by simulations.
	Author: Tao Jiang
	Email: tjiang@hit.edu.cn
"""

def parseArgs(argv):
	parser = argparse.ArgumentParser(prog="Trio_eval", description=USAGE, formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument("choice", type=str, help="Chose specific SV type.[IID/DUP/BND]")
	parser.add_argument("ans", type=str, help="Ground truth of simulations.")
	parser.add_argument("gt", type=str, help="Genotype in each chromosome.")
	parser.add_argument('cuteSV', type=str, help = "cuteSV callsets")
	parser.add_argument('sniffles', type=str, help = "Sniffles callsets")
	parser.add_argument('pbsv', type=str, help = "PBSV callsets")
	parser.add_argument('svim', type=str, help = "SVIM callsets")
	parser.add_argument('-b', '--bias', help = "Bias of overlaping.[%(default)s]", default = 0.7, type = float)
	parser.add_argument('-o', '--offect', help = "Offect of translocation overlaping.[%(default)s]", default = 1000, type = int)
	args = parser.parse_args(argv)
	return args

def setupLogging(debug=False):
	logLevel = logging.DEBUG if debug else logging.INFO
	logFormat = "%(asctime)s [%(levelname)s] %(message)s"
	logging.basicConfig( stream=sys.stderr, level=logLevel, format=logFormat )
	logging.info("Running %s" % " ".join(sys.argv))

if __name__ == '__main__':
	main(sys.argv[1:])


from cuteSV.cuteSV_genotype import cal_CIPOS, overlap_cover, assign_gt_fc
from multiprocessing import Pool
from pysam import VariantFile
import math
import time
import logging
import numpy as np
import pickle
from sklearn.cluster import KMeans

def parse_svtype(sv_type):
    if 'DEL' in sv_type:
        return 'DEL'
    if 'INS' in sv_type:
        return 'INS'
    if 'INV' in sv_type:
        return 'INV'
    if 'DUP' in sv_type:
        return 'DUP'
    if 'TRA' in sv_type:
        return 'TRA'
    if 'BND' in sv_type:
        return 'BND'
    return 'NA'

def parse_to_int(sth):
    if sth == None:
        return 0
    elif isinstance(sth, str):
        return int(sth)
    elif isinstance(sth, list):
        return parse_to_int(sth[0])
    elif isinstance(sth, tuple):
        return parse_to_int(sth[0])
    elif isinstance(sth, int):
        return sth
    else:
        return sth

# INDEL end->len, others:end
def parse_record(record):
    sv_type = parse_svtype(record.info['SVTYPE'])
    chrom1 = record.chrom
    start = parse_to_int(record.pos)
    chrom2 = record.chrom
    end = None
    ref = record.ref
    alts = record.alts[0]
    # read sv length from 1.SVLEN= 2.alt-ref 3.record as 0
    if 'SVLEN' in record.info:
        svlen = abs(parse_to_int(record.info['SVLEN']))
    elif alts[0] != '<' and (sv_type != 'TRA' and sv_type != 'BND'):
        svlen = abs(len(alts) - len(ref))
    else:
        svlen = 0
    # read end from 1.BND-alt 2.END= 3.start+svlen
    if sv_type == 'TRA' or sv_type == 'BND':
        try:
            tra_alt = str(record.alts[0])
            if tra_alt[0] == 'N':
                if tra_alt[1] == '[':
                    tra_type = 'A'
                else:
                    tra_type = 'B'
            elif tra_alt[0] == '[':
                tra_type = 'C'
            else:
                tra_type = 'D'
            if tra_alt[0] == 'N':
                tra_alt = tra_alt[2:-1]
            else:
                tra_alt = tra_alt[1:-2]
            if ':' in tra_alt:
                chrom2 = tra_alt.split(':')[0]
                end = int(tra_alt.split(':')[1])
        except:
            pass
    if end == None:
        if 'END' in record.info:
            end = parse_to_int(record.info['END'])
        else:
            end = start + svlen
    # chrom2 for BND
    if 'CHR2' in record.info:
        chrom2 = record.info['CHR2']
    # strand
    strand = '.'
    if 'STRAND' in record.info:
        strand = record.info['STRAND']  
    elif 'STRANDS' in record.info:
        strand = record.info['STRANDS']
    if isinstance(strand, tuple) or isinstance(strand, list):
        strand = strand[0]
    svid = record.id
    # update ref and alt with SEQ
    if 'SEQ' in record.info:
        if record.info['SVTYPE'] == 'INS' and record.alts[0] == '<INS>':
            alts = record.info['SEQ']
        if record.info['SVTYPE'] == 'DEL' and record.alts[0] == '<DEL>':
            ref = record.info['SEQ']
    return sv_type, chrom1, chrom2, start, end, svlen, strand, svid, ref, alts

def parse_sigs_chrom(var_type, work_dir, chrom_list, index):
    var_dict={}
    if False and var_type != 'TRA':
        with open(work_dir + var_type + '.pickle', 'rb') as f:
            for chrom in chrom_list:
                if chrom not in index[var_type].keys():
                    continue
                f.seek(index[var_type][chrom])
                var_dict[chrom]=pickle.load(f)
    else:
        with open(work_dir + var_type + '.pickle', 'rb') as f:
            for chrom in chrom_list:
                if chrom not in index[var_type].keys():
                    continue
                f.seek(index[var_type][chrom])
                sigs=pickle.load(f)
                #file_sig[]=mem_sig[DEL: -2,-1,0,1,2, INS: -2,-1,0,1,2,3, DUP: -2,-1,0,1,2, INV: -2,-1,0,1,2,3, TRA: -2,-1,0,1,2,3,4, reads: -1,0,1,2,3]
                #from file: DEL,DUP:  1,2,3,4, INS: 1,2,3,4,5, INV:1,3,4,5, TRA[1][4]:4,3,5,6
                #from mem: DEL,DUP: -1,0,1,2, INS: -1,0,1,2,3, INV: -1,1,2,3, TRA[-1][2]:2,1,3,4
                if var_type == 'DEL' or var_type == 'DUP':
                    for seq in sigs:
                        if chrom not in var_dict:
                            var_dict[chrom] = []
                        var_dict[chrom].append([seq[-1], int(seq[0]), int(seq[1]), seq[2]])
                elif var_type == 'INS':
                    for seq in sigs:
                        if chrom not in var_dict:
                            var_dict[chrom] = []
                        if len(seq) < 6:
                            cigar = '<INS>'
                        else:
                            cigar = seq[3]
                        cigar = '<INS>'
                        var_dict[chrom].append([seq[-1], int(seq[0]), int(seq[1]), seq[2], cigar])
                elif var_type == 'INV':
                    for seq in sigs:
                        if chrom not in var_dict:
                            var_dict[chrom] = []
                        var_dict[chrom].append([seq[-1], int(seq[1]), int(seq[2]), seq[3]])
                else:
                    for seq in sigs:
                        chrom1 = seq[-1]
                        tra_type = seq[0]
                        pos1 = int(seq[1])
                        chrom2 = seq[2]
                        pos2 = int(seq[3])
                        read_id = seq[4]
                        if chrom1 not in var_dict:
                            var_dict[chrom1] = dict()
                        if chrom2 not in var_dict[chrom1]:
                            var_dict[chrom1][chrom2] = []
                        var_dict[chrom1][chrom2].append([chrom2, pos1, pos2, read_id])
                    for chr1 in var_dict:
                        for chr2 in var_dict[chr1]:
                            var_dict[chr1][chr2].sort(key=lambda x:x[1])
            return var_dict

def check_same_variant(sv_type, end1, end2, bias):
    if sv_type == 'INS' or sv_type == 'DEL':
        return 0.7 < min(end1, end2) / max(end1, end2) <= 1
    return abs(end1 - end2) < bias

# var_list read_id which is similar to (pos, sv_end), the reads support the variant
def find_in_list(var_type, var_list, bias, pos, sv_end):
    if len(var_list) == 0:
        return [], pos, pos
    left = 0
    right = len(var_list) - 1
    mid = 0
    while left < right:
        mid = (left + right) >> 1
        if var_list[mid][1] < pos:
            left = mid + 1
        else:
            right = mid
    read_id_list = set()
    search_start = -1
    search_end = -1
    if right > 0 and pos - var_list[right - 1][1] <= bias: ##
        for i in range(right - 1, -1, -1):
            if check_same_variant(var_type, var_list[i][2], sv_end, bias):
                read_id_list.add(var_list[i][3])
                search_start = var_list[i][1]
            if i > 0 and (var_list[i][1] - var_list[i - 1][1] > bias or pos - var_list[i - 1][1] > bias):
                break
    if var_list[right][1] - pos <= bias:
        for i in range(right, len(var_list)):
            if check_same_variant(var_type, var_list[i][2], sv_end, bias):  # if abs(var_list[i][2] - sv_end) < 1000:
                read_id_list.add(var_list[i][3])
                search_end = var_list[i][1]
            if i < len(var_list) - 1 and (var_list[i + 1][1] - var_list[i][1] > bias or var_list[i + 1][1] - pos > bias):
                break
    if search_start == -1:
        search_start = pos
    if search_end == -1:
        search_end = pos
    search_threshold = max(abs(pos - search_start), abs(pos - search_end))
    if search_start > search_end:
        search_start, search_end = search_end, search_start
    if search_start == search_end:
        search_end += 1
    return list(read_id_list), search_start, search_end

def find_in_indel_list(var_type, var_list, bias_origin, pos, sv_end, threshold_gloab, multi_allele):
    # bias = min(bias_origin * 10, 2000)
    bias = bias_origin
    debug_pos = -1
    if len(var_list) == 0:
        return [], pos, pos, '.,.', '.,.'
    left = 0
    right = len(var_list) - 1
    mid = 0
    while left < right:
        mid = (left + right) >> 1
        if var_list[mid][1] < pos:
            left = mid + 1
        else:
            right = mid
    candidates = []
    if right > 0 and pos - var_list[right - 1][1] <= bias:
        for i in range(right - 1, -1, -1):
            candidates.append(var_list[i]) # [chrom, start, len, read_id] or [chrom, start, len, read_id, seq]
            if i > 0 and (var_list[i][1] - var_list[i - 1][1] > bias or pos - var_list[i - 1][1] > 2 * bias):
                break
    if var_list[right][1] - pos <= bias:
        for i in range(right, len(var_list)):
            candidates.append(var_list[i])
            if i < len(var_list) - 1 and (var_list[i + 1][1] - var_list[i][1] > bias or var_list[i + 1][1] - pos > 2 * bias):
                break
    if len(candidates) == 0:
        return [], pos, pos, '.,.', '.,.'
    read_tag = dict()
    for element in candidates:
        if element[3] not in read_tag:
            read_tag[element[3]] = []
        read_tag[element[3]].append(element)

    # merge sigs on the same read
    read_tag2SortedList = []

    for read_id in read_tag: # 40
        for i in range(len(read_tag[read_id])):
            read_tag2SortedList.append(read_tag[read_id][i])
            if i + 1 < len(read_tag[read_id]):
                j = i + 1
                if var_type == 'DEL':
                    read_tag2SortedList.append(
                        [read_tag[read_id][i][0], int((read_tag[read_id][i][1] + read_tag[read_id][j][1]) / 2),
                         read_tag[read_id][i][2] + read_tag[read_id][j][2], read_tag[read_id][i][3]])
                else:
                    read_tag2SortedList.append(
                        [read_tag[read_id][i][0], int((read_tag[read_id][i][1] + read_tag[read_id][j][1]) / 2),
                         read_tag[read_id][i][2] + read_tag[read_id][j][2], read_tag[read_id][i][3],
                         read_tag[read_id][i][4]])
                if j + 1 < len(read_tag[read_id]):
                    k = j + 1
                    if var_type == 'DEL':
                        read_tag2SortedList.append([read_tag[read_id][i][0], int((read_tag[read_id][i][1] +
                                                                                read_tag[read_id][j][1] +
                                                                                read_tag[read_id][k][1]) / 3),
                                                    read_tag[read_id][i][2] + read_tag[read_id][j][2] +
                                                    read_tag[read_id][k][2], read_tag[read_id][i][3]])
                    else:
                        read_tag2SortedList.append([read_tag[read_id][i][0], int((read_tag[read_id][i][1] +
                                                                                read_tag[read_id][j][1] +
                                                                                read_tag[read_id][k][1]) / 3),
                                                    read_tag[read_id][i][2] + read_tag[read_id][j][2] +
                                                    read_tag[read_id][k][2], read_tag[read_id][i][3],
                                                    read_tag[read_id][i][4]])

    read_tag2SortedList = sorted(read_tag2SortedList, key = lambda x:x[2])
    if pos == debug_pos:
        print(read_tag2SortedList)
    global_len = [i[2] for i in read_tag2SortedList]
    DISCRETE_THRESHOLD_LEN_CLUSTER_TEMP = threshold_gloab * np.mean(global_len)
    if var_type == 'DEL':
        last_len = read_tag2SortedList[0][2]
        cur_bias = last_len * threshold_gloab
        allele_collect = list()
        allele_collect.append([[read_tag2SortedList[0][1]],  # start
                                [read_tag2SortedList[0][2]],  # len
                                [],  # support
                                [read_tag2SortedList[0][3]]])  # read_id
        for i in read_tag2SortedList[1:]:
            if i[2] - last_len > cur_bias:
                allele_collect[-1][2].append(len(allele_collect[-1][0]))
                allele_collect.append([[],[],[],[]])
            allele_collect[-1][0].append(i[1])
            allele_collect[-1][1].append(i[2])
            allele_collect[-1][3].append(i[3])
            last_len = (last_len * (len(allele_collect[-1][0]) - 1) + i[2]) / len(allele_collect[-1][0])
            cur_bias = last_len * threshold_gloab
        if pos == debug_pos:
            print(allele_collect)
            print('above is all allele collect')
        allele_collect[-1][2].append(len(allele_collect[-1][0]))
        allele_idx = -1
        nearest_gap = 0x3f3f3f3f
        for i in range(len(allele_collect)):
            allele = allele_collect[i]
            signalLen = np.mean(allele[1])
            if min(signalLen, sv_end) / max(signalLen, sv_end) > 0.7:
                if abs(signalLen - sv_end) < nearest_gap:
                    allele_idx = i
                    nearest_gap = abs(signalLen - sv_end)
        if allele_idx == -1:
            lower_length = sv_end * 0.7
            upper_length = sv_end / 0.7
            final_alleles = [[],[],[],[]]
            for i in range(len(allele_collect[allele_idx][0])):
                if lower_length <= allele_collect[allele_idx][1][i] <= upper_length:
                    final_alleles[0].append(allele_collect[allele_idx][0][i])
                    final_alleles[1].append(allele_collect[allele_idx][1][i])
                    final_alleles[3].append(allele_collect[allele_idx][3][i])
        else:
            final_alleles = allele_collect[allele_idx]

        if pos == debug_pos:
            print(allele_idx)
            print(final_alleles[:2])
        if multi_allele:
            bimodal_data = np.array(final_alleles[1])
            if len(bimodal_data) > 1 and final_alleles[1][0] != final_alleles[1][-1]:
                model = KMeans(n_clusters=2, init=np.array([int(len(bimodal_data)/4), int(len(bimodal_data)/4*3)]).reshape(-1, 1), n_init=1)
                # try:
                model.fit(bimodal_data.reshape(-1, 1))
                # except:
                #     print('reshape error')
                labels = model.labels_
                if pos == debug_pos:
                    print(labels)
                cate = 0
                for i in range(0, len(labels) - 1):
                    if labels[i] != labels[i + 1]:
                        cate = i + 1
                        break
                delta0 = math.ceil(cate / 8) if cate >= 3 else 0
                delta1 = math.ceil((len(labels) - cate + 1) / 8) if len(labels) - cate >= 3 else 0
                min_alleles = [final_alleles[1][delta0], final_alleles[1][cate + delta1]]
                max_alleles = [final_alleles[1][cate - delta0 - 1], final_alleles[1][len(labels) - delta1 - 1]]
                final_alleles_filter = [[],[],[],[]]
                if abs(max_alleles[0]-max_alleles[1]) >= max(3*max(max_alleles[0]-min_alleles[0], max_alleles[1]-min_alleles[1]), 6):
                    allele0 = np.mean(final_alleles[1][delta0:(cate-delta0)])
                    allele1 = np.mean(final_alleles[1][cate+delta1:]) if delta1 == 0 else np.mean(final_alleles[1][cate+delta1:-delta1])
                    if min(allele0, sv_end) / max(allele0, sv_end) >= min(allele1, sv_end) / max(allele1, sv_end): # choose front
                        if min(min_alleles[0], sv_end) / max(min_alleles[0], sv_end) > 0.9 and min(max_alleles[0], sv_end) / max(max_alleles[0], sv_end) > 0.9:
                            if cate >= max(3, len(labels) / 5):
                                for i in range(cate):
                                    # if min(final_alleles[1][i], sv_end) / max(final_alleles[1][i], sv_end) > 0.9:
                                    for j in [0, 1, 3]:
                                        final_alleles_filter[j].append(final_alleles[j][i])
                    elif min(min_alleles[1], sv_end) / max(min_alleles[1], sv_end) > 0.9 and min(max_alleles[1], sv_end) / max(max_alleles[1], sv_end) > 0.9:
                        if len(labels) - cate >= max(3, len(labels) / 5):
                            for i in range(cate, len(labels)):
                                # if min(final_alleles[1][i], sv_end) / max(final_alleles[1][i], sv_end) > 0.9:
                                for j in [0, 1, 3]:
                                    final_alleles_filter[j].append(final_alleles[j][i])
                if len(final_alleles_filter[0]) > 0:
                    final_alleles = final_alleles_filter
                if pos == debug_pos:
                    print('final alleles:')
                    print(final_alleles[:2])
        if len(final_alleles[3]) > 0:
            read_id_set = set(final_alleles[3])
            CIPOS = cal_CIPOS(np.std(final_alleles[0]), len(final_alleles[0]))
            CILEN = cal_CIPOS(np.std(final_alleles[1]), len(final_alleles[1]))
            seq = '<DEL>'
            # search_start = np.median(final_alleles[0])
            # search_end = np.median(final_alleles[0])
            search_start = min(final_alleles[0])
            search_end = max(final_alleles[0])
            search_threshold = min(abs(pos - search_start), abs(pos - search_end))
            if pos == debug_pos:
                print('search_start:%f'%(search_start))
                print(final_alleles[0])
        else:
            read_id_set = set()
            CIPOS = "-0,0"
            CILEN = "-0,0"
            seq = '<DEL>'
            search_start = pos
            search_end = pos
            search_threshold = 0
    if var_type == 'INS':
        last_len = read_tag2SortedList[0][2]
        cur_bias = last_len * threshold_gloab
        allele_collect = list()
        allele_collect.append([[read_tag2SortedList[0][1]],  # start
                                [read_tag2SortedList[0][2]],  # len
                                [],  # support
                                [read_tag2SortedList[0][3]],  # read_id
                                [read_tag2SortedList[0][4]]])  # ins_seq
        for i in read_tag2SortedList[1:]:
            if i[2] - last_len > cur_bias:
                allele_collect[-1][2].append(len(allele_collect[-1][0]))
                allele_collect.append([[],[],[],[],[]])
            allele_collect[-1][0].append(i[1])
            allele_collect[-1][1].append(i[2])
            allele_collect[-1][3].append(i[3])
            allele_collect[-1][4].append(i[4])
            last_len = (last_len * (len(allele_collect[-1][0]) - 1) + i[2]) / len(allele_collect[-1][0])
            cur_bias = last_len * threshold_gloab
        if pos == debug_pos:
            print(allele_collect)
            print('above is all allele collect')
        allele_collect[-1][2].append(len(allele_collect[-1][0]))
        allele_idx = -1
        nearest_gap = 0x3f3f3f3f
        for i in range(len(allele_collect)):
            allele = allele_collect[i]
            signalLen = np.mean(allele[1])
            if min(signalLen, sv_end) / max(signalLen, sv_end) > 0.7:
                if abs(signalLen - sv_end) < nearest_gap:
                    allele_idx = i
                    nearest_gap = abs(signalLen - sv_end)
        if allele_idx == -1:
            lower_length = sv_end * 0.7
            upper_length = sv_end / 0.7
            final_alleles = [[],[],[],[]]
            for i in range(len(allele_collect[allele_idx][0])):
                if lower_length <= allele_collect[allele_idx][1][i] <= upper_length:
                    final_alleles[0].append(allele_collect[allele_idx][0][i])
                    final_alleles[1].append(allele_collect[allele_idx][1][i])
                    final_alleles[3].append(allele_collect[allele_idx][3][i])
                    # final_alleles[4].append(allele_collect[allele_idx][4][i])
        else:
            final_alleles = allele_collect[allele_idx]

        if pos == debug_pos:
            print(allele_idx)
            print(final_alleles[:2])
        if multi_allele:
            bimodal_data = np.array(final_alleles[1])
            if len(bimodal_data) > 1 and final_alleles[1][0] != final_alleles[1][-1]:
                model = KMeans(n_clusters=2, init=np.array([int(len(bimodal_data)/4), int(len(bimodal_data)/4*3)]).reshape(-1, 1), n_init=1)
                # try:
                model.fit(bimodal_data.reshape(-1, 1))
                # except:
                #     print('reshape error')
                labels = model.labels_
                if pos == debug_pos:
                    print(labels)
                cate = 0
                for i in range(0, len(labels) - 1):
                    if labels[i] != labels[i + 1]:
                        cate = i + 1
                        break
                delta0 = math.ceil(cate / 8) if cate >= 5 else 0
                delta1 = math.ceil((len(labels) - cate) / 8) if len(labels) - cate >= 5 else 0
                min_alleles = [final_alleles[1][delta0], final_alleles[1][cate + delta1]]
                max_alleles = [final_alleles[1][cate - delta0 - 1], final_alleles[1][len(labels) - delta1 - 1]]
                final_alleles_filter = [[],[],[],[]]
                if abs(max_alleles[0]-max_alleles[1]) >= max(3*max(max_alleles[0]-min_alleles[0], max_alleles[1]-min_alleles[1]), 6): # or > 3
                    allele0 = np.mean(final_alleles[1][delta0:(cate-delta0)])
                    allele1 = np.mean(final_alleles[1][cate+delta1:]) if delta1 == 0 else np.mean(final_alleles[1][cate+delta1:-delta1])
                    if min(allele0, sv_end) / max(allele0, sv_end) >= min(allele1, sv_end) / max(allele1, sv_end): # choose front
                        if min(min_alleles[0], sv_end) / max(min_alleles[0], sv_end) > 0.9 and min(max_alleles[0], sv_end) / max(max_alleles[0], sv_end) > 0.9:
                            if cate >= max(3, len(labels) / 5):
                                for i in range(cate):
                                    # if min(final_alleles[1][i], sv_end) / max(final_alleles[1][i], sv_end) > 0.9:
                                    for j in [0, 1, 3]:
                                        final_alleles_filter[j].append(final_alleles[j][i])
                    elif min(min_alleles[1], sv_end) / max(min_alleles[1], sv_end) > 0.9 and min(max_alleles[1], sv_end) / max(max_alleles[1], sv_end) > 0.9:
                        if len(labels) - cate >= max(3, len(labels) / 5):
                            for i in range(cate, len(labels)):
                                # if min(final_alleles[1][i], sv_end) / max(final_alleles[1][i], sv_end) > 0.9:
                                for j in [0, 1, 3]:
                                    final_alleles_filter[j].append(final_alleles[j][i])
                if len(final_alleles_filter[0]) > 0:
                    final_alleles = final_alleles_filter
                if pos == debug_pos:
                    print('final alleles:')
                    print(final_alleles[:2])
        if len(final_alleles[3]) > 0:
            read_id_set = set(final_alleles[3])
            CIPOS = cal_CIPOS(np.std(final_alleles[0]), len(final_alleles[0]))
            CILEN = cal_CIPOS(np.std(final_alleles[1]), len(final_alleles[1]))
            seq = '<INS>'
            # search_start = np.median(final_alleles[0])
            # search_end = np.median(final_alleles[0])
            search_start = min(final_alleles[0])
            search_end = max(final_alleles[0])
            search_threshold = min(abs(pos - search_start), abs(pos - search_end))
        else:
            read_id_set = set()
            CIPOS = "-0,0"
            CILEN = "-0,0"
            seq = '<INS>'
            search_start = pos
            search_end = pos
            search_threshold = 0
    # return list(read_id_set), search_threshold, CIPOS, CILEN
    return list(read_id_set), search_start, search_end, CIPOS, CILEN

def generate_dispatch(reads_count, chrom_list):
    dispatch = [[]]
    cur_count = 0
    for item in reads_count:
        if cur_count >= 10000:
            dispatch.append([])
            cur_count = 0
        cur_count = item[1]
        dispatch[-1].append(item[0])
    # chrom without reads
    for chrom in chrom_list:
        flag = 0
        for x in reads_count:
            if chrom == x[0]:
                flag = 1
        if flag == 0:
            dispatch[0].append(chrom)
    return dispatch


def force_calling_chrom(ivcf_path, temporary_dir, max_cluster_bias_dict, threshold_gloab_dict, gt_round, read_range, threads, sigs_index):
    logging.info('Check the parameter -Ivcf: OK.')
    logging.info('Enable to perform force calling.')
    if sigs_index==None:
        with open("%s/sigindex.pickle"%temporary_dir,"rb") as f:
            sigs_index=pickle.load(f)
    # parse svs tobe genotyped
    vcf_reader = VariantFile(ivcf_path, 'r')
    svs_tobe_genotyped = dict()
    svs_pre = dict()
    svs_multi = dict()
    for record in vcf_reader.fetch():
        sv_type, chrom, sv_chr2, pos, sv_end, sv_len, sv_strand, svid, ref, alts = parse_record(record)
        if sv_type not in ["DEL", "INS", "DUP", "INV", "TRA", "BND"]:
            continue
        if chrom not in svs_tobe_genotyped:
            svs_tobe_genotyped[chrom] = list()
            svs_pre[chrom] = dict()
        svs_tobe_genotyped[chrom].append([sv_type, sv_chr2, pos, sv_end, sv_len, svid, ref, alts, sv_strand, chrom])
        if pos not in svs_pre[chrom]:
            svs_pre[chrom][pos] = 0
        svs_pre[chrom][pos] += 1
    for c in svs_pre:
        for s in svs_pre[c]:
            if svs_pre[c][s] == 2:
                if c not in svs_multi:
                    svs_multi[c] = set()
                svs_multi[c].add(s)
    start_time = time.time()
    # parse reads in alignment
    reads_count = sigs_index["reads_count"]
    reads_count = sorted(reads_count.items(), key=lambda x:x[1])
    dispatch = generate_dispatch(reads_count, svs_tobe_genotyped.keys())
    logging.info('finish dispatch in {}.'.format(time.time() - start_time))

    # force calling
    pool_result = list()
    result= {}
    process_pool = Pool(processes = threads)
    # dispatch = [['MT']]
    for chroms in dispatch:
        genotype_sv_list = dict()
        for chrom in chroms:
            if chrom in svs_tobe_genotyped:
                genotype_sv_list[chrom] = svs_tobe_genotyped[chrom]
        if len(genotype_sv_list) == 0:
            continue
        fx_para = [(chroms, genotype_sv_list, temporary_dir, max_cluster_bias_dict, threshold_gloab_dict, gt_round, sigs_index, read_range, svs_multi)]
        pool_result.append(process_pool.map_async(solve_fc_wrapper, fx_para))
    process_pool.close()
    process_pool.join()

    for x in pool_result:
        result.update(x.get()[0])
    return result

def solve_fc_wrapper(args):
    return solve_fc(*args)
def solve_fc(chrom_list, svs_dict, temporary_dir, max_cluster_bias_dict, threshold_gloab_dict, gt_round, sigs_index, read_range, svs_multi):
    reads_info = dict() # [10000, 10468, 0, 'm54238_180901_011437/52298335/ccs']
    readsfile = open("%sreads.pickle"%(temporary_dir), 'rb')
    for chrom in chrom_list:
        readsfile.seek(sigs_index["reads"][chrom])
        reads_info[chrom]=pickle.load(readsfile)
    readsfile.close()
    sv_dict = dict()
    for sv_type in ["DEL", "DUP", "INS", "INV", "TRA"]:
        sv_dict[sv_type] = parse_sigs_chrom(sv_type, temporary_dir, chrom_list, sigs_index)
    
    gt_list = {}
    for chrom in svs_dict:
        gt_list[chrom]=[]
        start_time = time.time()
        read_id_dict = dict()
        svtype_id_dict = dict()
        ci_dict = dict()
        search_list = list()
        for i in range(len(svs_dict[chrom])):
            record = svs_dict[chrom][i]
            # [sv_type, sv_chr2, pos, sv_end, sv_len, svid, ref, alts, sv_strand, chrom]
            sv_type = record[0]
            sv_chr2 = record[1]
            sv_start = record[2]
            sv_end = record[3]
            sv_len = record[4]
            chrom = record[9]
            search_id_list = list()
            # rewrite!
            if (sv_type == 'TRA' or sv_type == 'BND') and 'TRA' in sv_dict and chrom in sv_dict['TRA'] and sv_chr2 in sv_dict['TRA'][chrom]:
                search_id_list = sv_dict['TRA'][chrom][sv_chr2]
            elif sv_type != 'TRA' and sv_type in sv_dict and chrom in sv_dict[sv_type]:
                search_id_list = sv_dict[sv_type][chrom]
            max_cluster_bias = 0
            if sv_type == 'INS' or sv_type == 'DEL':
                sigs_bias = max_cluster_bias_dict[sv_type]
                if chrom in svs_multi and sv_start in svs_multi[chrom]:
                    multi_allele = True
                else:
                    multi_allele = False
                read_id_list, search_start, search_end, CIPOS, CILEN = find_in_indel_list(sv_type, search_id_list, sigs_bias, sv_start, sv_len, threshold_gloab_dict[sv_type], multi_allele)
            else:
                sv_temp = sv_type
                sigs_bias = max_cluster_bias_dict[sv_type if sv_type != 'BND' else 'TRA']
                if sv_len / 2 > sigs_bias:
                    sigs_bias = sv_len / 2
                # read_id_list, max_cluster_bias = find_in_list(sv_type, search_id_list, sigs_bias, sv_start, sv_end)
                read_id_list, search_start, search_end = find_in_list(sv_type, search_id_list, sigs_bias, sv_start, sv_end)
                CIPOS = '.'
                CILEN = '.'
            max_cluster_bias = max(abs(sv_start - search_start), abs(sv_start - search_end))
            max_cluster_bias = max(read_range, max_cluster_bias)
            if sv_type == 'INS' or sv_type == 'TRA' or sv_type == 'BND':
                search_list.append((max(sv_start - max_cluster_bias, 0), sv_start + max_cluster_bias))
            elif sv_type == 'DEL':
                if read_range < 500:
                    search_list.append((max(sv_start - max_cluster_bias, 0), sv_start + max_cluster_bias))
                else:
                    search_list.append((max(sv_start + abs(sv_len) / 5, 0), sv_start + abs(sv_len) - abs(sv_len) / 5))
            elif sv_type == 'INV':
                search_list.append((search_start, search_end+1))
            elif sv_type == 'DUP':
                search_list.append((sv_start, sv_end))

            read_id_dict[i] = read_id_list
            svtype_id_dict[i] = sv_type
            ci_dict[i] = (CIPOS, CILEN)
        if chrom in reads_info:
            iteration_dict, primary_num_dict, cover_dict, overlap_dict = overlap_cover(search_list, reads_info[chrom]) # both key(sv idx), value(set(read id))
        else:
            iteration_dict = dict()
            primary_num_dict = dict()
            cover_dict = dict()
            overlap_dict = dict()
            for i in read_id_dict:
                iteration_dict[i] = 0
                primary_num_dict[i] = 0
                cover_dict[i] = set()
                overlap_dict[i] = set()
        assert len(iteration_dict) == len(read_id_dict), "overlap length error"
        assert len(cover_dict) == len(svs_dict[chrom]), "cover length error"
        assert len(overlap_dict) == len(svs_dict[chrom]), "overlap length error"
        assign_list = assign_gt_fc(iteration_dict, primary_num_dict, cover_dict, overlap_dict, read_id_dict, svtype_id_dict)
        for i in range(len(svs_dict[chrom])):
            assert len(assign_list[i]) == 6, "assign genotype error"
            record = svs_dict[chrom][i]
            rname = ','.join(read_id_dict[i])
            if rname == '':
                rname = 'NULL'
            if record[7] == '<TRA>' or record[7] == '<BND>':
                seq = str(record[1]) + ':' + str(record[3])
            else:
                seq = '<' + record[0] + '>'
            # [sv_type, sv_chr2, pos, sv_end, sv_len, svid, ref, alts, sv_strand, chrom]
            gt_list[record[9]].append([record[9], record[2], assign_list[i][2], record[0], record[3],
                        ci_dict[i][0], ci_dict[i][1], assign_list[i], rname, record[5],
                        record[6], record[7],
                        record[8], seq, record[4]])
        logging.info("Finished calling %s."%(chrom))
    return gt_list



import sys
import argparse
import logging
import time

def population_statistic(pop_merged_vcf, output_file):
    output = open(output_file, 'w')
    idx = 0
    with open(pop_merged_vcf, 'r') as f:
        for line in f:
            if line[0] == '#':
                continue
            seq = line.strip().split('\t')
            svlen = abs(int(seq[7].split(';SVLEN=')[1].split(';')[0]))
            svtype = seq[7].split(';SVTYPE=')[1].split(';')[0]
            if (svtype != 'TRA' and svtype != 'BND') and svlen < 50:
                continue
            idx += 1
            af = float(seq[7].split(';AF=')[1].split(';')[0])
            hwe = float(seq[7].split(';HWE=')[1].split(';')[0])
            exchet = float(seq[7].split(';ExcHet=')[1])
            missing_cnt = 0
            for i in range(9, 109, 1):
                if seq[i][0] == '.':
                    missing_cnt += 1
                if seq[i][2] == '.':
                    missing_cnt += 1
            output.write('%d\t%f\t%f\t%f\t%f\n'%(idx, missing_cnt/200, af, hwe, exchet))

def compare_callsets(pop_vcf_file, base_vcf_file, output_file):
    def parse(file):
        svs_dict = dict()
        with open(file, 'r') as f:
            for line in f:
                if line[0] == '#':
                    continue
                seq = line.strip().split('\t')
                chrom = seq[0]
                pos = int(seq[1])
                svtype = seq[7].split('SVTYPE=')[1].split(';')[0]
                svlen = abs(int(seq[7].split(';SVLEN=')[1].split(';')[0]))
                if (svtype != 'TRA' and svtype != 'BND') and svlen < 50:
                    continue
                hwe = float(seq[7].split(';HWE=')[1].split(';')[0])
                exchet = float(seq[7].split(';ExcHet=')[1])
                af = float(seq[7].split(';AF=')[1].split(';')[0])
                missing_cnt = 0
                for gt in seq[9:]:
                    if gt[0] == '.':
                        missing_cnt += 1
                    if gt[2] == '.':
                        missing_cnt += 1
                if missing_cnt > 10 or hwe < 0.000001 or exchet < 0.000001:
                    continue
                if svtype == 'DEL' or svtype == 'INS':
                    svlen = abs(int(seq[7].split('SVLEN=')[1].split(';')[0]))
                    af = float(seq[7].split(';AF=')[1].split(';')[0])
                    if chrom not in svs_dict:
                        svs_dict[chrom] = list()
                    svs_dict[chrom].append([pos, svtype, svlen, af])
        return svs_dict
    def parse_base(file):
        svs_dict = dict()
        with open(file, 'r') as f:
            for line in f:
                if line[0] == '#':
                    continue
                seq = line.strip().split('\t')
                chrom = seq[0]
                pos = int(seq[1])
                svtype = seq[7].split('SVTYPE=')[1].split(';')[0]
                af = float(seq[7].split(';AF=')[1].split(';')[0])
                if svtype == 'DEL' or svtype == 'INS':
                    svlen = abs(int(seq[7].split('SVLEN=')[1].split(';')[0]))
                    if chrom not in svs_dict:
                        svs_dict[chrom] = list()
                    svs_dict[chrom].append([pos, svtype, svlen, af])
        return svs_dict
    base = parse_base(base_vcf_file)
    comp = parse(pop_vcf_file)
    output = open(output_file, 'w')
    pos_bias = 1000
    length_ratio = 0.7
    for chrom in base:
        if chrom in comp:
            for basesv in base[chrom]:
                for compsv in comp[chrom]:
                    if basesv[1] == compsv[1] and abs(basesv[0] - compsv[0]) <= pos_bias and min(basesv[2], compsv[2]) / max(basesv[2], compsv[2]) > length_ratio:
                        output.write('%s\t%f\t%f\t%f\n'%(basesv[1], basesv[3], compsv[3], basesv[3]-compsv[3]))
                        break
    output.close()

def pre_cmrg(input_vcf, output_vcf):
    # input_vcf -> HG002_GRCh37_CMRG_SV_v1.00.vcf
    output = open(output_vcf, 'w')
    with open(input_vcf, 'r') as f:
        for line in f:
            if line[0] == '#':
                if line[1] != '#':
                    output.write('##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">\n')
                    output.write('##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Difference in length between REF and ALT alleles">\n')
                output.write(line)
            else:
                seq = line.strip().split('\t')
                ref = seq[3]
                alt = seq[4]
                for i in range(7):
                    output.write(seq[i] + '\t')
                if len(ref) > len(alt): # DEL
                    output.write('SVTYPE=DEL;SVLEN=%d'%(len(alt) - len(ref)))
                else: # INS
                    output.write('SVTYPE=INS;SVLEN=%d'%(len(alt) - len(ref)))
                output.write('\t%s\t%s\n'%(seq[8], seq[9]))
    output.close()

def main(argv):
    args = parseArgs(argv)
    setupLogging(False)
    starttime = time.time()
    if args.handle == 'POP':
        population_statistic(args.input, args.output)
    if args.handle == 'COMP':
        compare_callsets(args.input, args.base_vcf, args.output)
    if args.handel == 'CMRG':
        pre_cmrg(args.input, args.output)
    logging.info("Finished in %0.2f seconds."%(time.time() - starttime))

USAGE="""\
    Processing and evaluation of force calling.
    Author: Shuqi Cao
    Email: sqcao@stu.hit.edu.cn
"""

def parseArgs(argv):
    parser = argparse.ArgumentParser(prog="evaluation on population Statistics/callsets Compare worldwide cohort", description=USAGE, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("handle", type=str, help="The aspect of evaluation, contains CMRG/POP/COMP.")
    parser.add_argument("--input", type=str, help="Input VCF, including vcf file to be preprocessed or merged population callsets from force calling methods.")
    parser.add_argument("--base_vcf", type=str, help="Worldwide population callsets.")
    parser.add_argument("--output", type=str, help="Output file.")
    args = parser.parse_args(argv)
    return args

def setupLogging(debug=False):
    logLevel = logging.DEBUG if debug else logging.INFO
    logFormat = "%(asctime)s [%(levelname)s] %(message)s"
    logging.basicConfig( stream=sys.stderr, level=logLevel, format=logFormat )
    logging.info("Running %s" % " ".join(sys.argv))

if __name__ == '__main__':
    main(sys.argv[1:])



''' 
 * All rights Reserved, Designed By HIT-Bioinformatics   
 * @Title:  cuteSV_Description.py
 * @author: tjiang & sqcao & zdzhang
 * @date: Nov. 29th 2023
 * @version V2.1.0
'''
import argparse

VERSION = '2.1.0'

class cuteSVdp(object):
	'''
	Detailed descriptions of cuteSV version and its parameters.
	'''

	USAGE="""\
		
	Current version: v%s
	Author: Tao Jiang
	Contact: tjiang@hit.edu.cn

	If you use cuteSV in your work, please cite:
		Jiang T et al. Long-read-based human genomic structural variation detection with cuteSV. 
		Genome Biol 21,189(2020). https://doi.org/10.1186/s13059-020-02107-y


	Suggestions:

	For PacBio CLR data:
		--max_cluster_bias_INS		100
		--diff_ratio_merging_INS	0.3
		--max_cluster_bias_DEL	200
		--diff_ratio_merging_DEL	0.5

	For PacBio CCS(HIFI) data:
		--max_cluster_bias_INS		1000
		--diff_ratio_merging_INS	0.9
		--max_cluster_bias_DEL	1000
		--diff_ratio_merging_DEL	0.5

	For ONT data:
		--max_cluster_bias_INS		100
		--diff_ratio_merging_INS	0.3
		--max_cluster_bias_DEL	100
		--diff_ratio_merging_DEL	0.3


	"""%(VERSION)

	# MinSizeDel = 'For current version of cuteSV, it can detect deletions larger than this size.'

def parseArgs(argv):
	parser = argparse.ArgumentParser(prog="cuteSV", 
		description=cuteSVdp.USAGE, 
		formatter_class=argparse.RawDescriptionHelpFormatter)

	parser.add_argument('--version', '-v', 
		action = 'version', 
		version = '%(prog)s {version}'.format(version=VERSION))

	# **************Parameters of input******************
	parser.add_argument("input", 
		metavar="[BAM]", 
		type = str, 
		help ="Sorted .bam file from NGMLR or Minimap2.")
	parser.add_argument("reference",  
		type = str, 
		help ="The reference genome in fasta format.")
	parser.add_argument('output', 
		type = str, 
		help = "Output VCF format file.")
	parser.add_argument('work_dir', 
		type = str, 
		help = "Work-directory for distributed jobs")

	# ************** Other Parameters******************
	parser.add_argument('-t', '--threads', 
		help = "Number of threads to use.[%(default)s]", 
		default = 16, 
		type = int)
	parser.add_argument('-b', '--batches', 
		help = "Batch of genome segmentation interval.[%(default)s]", 
		default = 10000000, 
		type = int)
	# The description of batches needs to improve.
	parser.add_argument('-S', '--sample',
		help = "Sample name/id",
		default = "NULL",
		type = str)

	parser.add_argument('--retain_work_dir',
		help = "Enable to retain temporary folder and files.",
		action="store_true")
	
	parser.add_argument('--write_old_sigs',
		help = "Enable to write sigs file in temporary folder for legacy compatibilities.",
		action="store_true")

	parser.add_argument('--report_readid',
		help = "Enable to report supporting read ids for each SV.",
		action="store_true")

	# **************Parameters in signatures collection******************
	GroupSignaturesCollect = parser.add_argument_group('Collection of SV signatures')
	GroupSignaturesCollect.add_argument('-p', '--max_split_parts', 
		help = "Maximum number of split segments a read may be aligned before it is ignored. All split segments are considered when using -1. \
			(Recommand -1 when applying assembly-based alignment.)[%(default)s]", 
		default = 7, 
		type = int)
	GroupSignaturesCollect.add_argument('-q', '--min_mapq', 
		help = "Minimum mapping quality value of alignment to be taken into account.[%(default)s]", 
		default = 20, 
		type = int)
	GroupSignaturesCollect.add_argument('-r', '--min_read_len', 
		help = "Ignores reads that only report alignments with not longer than bp.[%(default)s]", 
		default = 500, 
		type = int)
	GroupSignaturesCollect.add_argument('-md', '--merge_del_threshold', 
		help = "Maximum distance of deletion signals to be merged. In our paper, I used -md 500 to process HG002 real human sample data.[%(default)s]", 
		default = 0, 
		type = int)
	GroupSignaturesCollect.add_argument('-mi', '--merge_ins_threshold', 
		help = "Maximum distance of insertion signals to be merged. In our paper, I used -mi 500 to process HG002 real human sample data.[%(default)s]", 
		default = 100, 
		type = int)
	GroupSignaturesCollect.add_argument('-include_bed', 
		help = "Optional given bed file. Only detect SVs in regions in the BED file. [NULL]",
		default = None,
        type = str)
	# The min_read_len in last version is 2000.
	# signatures with overlap need to be filtered

	# **************Parameters in clustering******************
	GroupSVCluster = parser.add_argument_group('Generation of SV clusters')
	GroupSVCluster.add_argument('-s', '--min_support', 
		help = "Minimum number of reads that support a SV to be reported.[%(default)s]", 
		default = 10, 
		type = int)
	GroupSVCluster.add_argument('-l', '--min_size', 
		help = "Minimum size of SV to be reported.[%(default)s]", 
		default = 30, 
		type = int)
	GroupSVCluster.add_argument('-L', '--max_size', 
		help = "Maximum size of SV to be reported. All SVs are reported when using -1. [%(default)s]", 
		default = 100000, 
		type = int)
	GroupSVCluster.add_argument('-sl', '--min_siglength', 
		help = "Minimum length of SV signal to be extracted.[%(default)s]", 
		default = 10, 
		type = int)

	# **************Parameters in genotyping******************
	GroupGenotype = parser.add_argument_group('Computing genotypes')
	GroupGenotype.add_argument('--genotype',
		help = "Enable to generate genotypes.",
		action="store_true")
	GroupGenotype.add_argument('--gt_round', 
		help = "Maximum round of iteration for alignments searching if perform genotyping.[%(default)s]", 
		default = 500, 
		type = int)
	GroupGenotype.add_argument('--read_range', 
		help = "The interval range for counting reads distribution.[%(default)s]", 
		default = 1000, 
		type = int)
	# GroupGenotype.add_argument('--hom', 
	# 	help = "Threshold on allele frequency for homozygous.[%(default)s]", 
	# 	default = 0.8, 
	# 	type = float)
	# GroupGenotype.add_argument('--het', 
	# 	help = "Threshold on allele frequency for heterozygous.[%(default)s].", 
	# 	default = 0.2, 
	# 	type = float)

	# Just a parameter for debug.
	# Will be removed in future.
	# GroupSVCluster.add_argument('--preset',
	# 	help = "Parameter presets for different sequencing technologies (pbclr/pbccs/ont).[%(default)s]",
	# 	default = "pbccs",
	# 	type = str)

	# **************Parameters in force calling******************
	GroupGenotype = parser.add_argument_group('Force calling')
	GroupGenotype.add_argument('-Ivcf', #'--MERGED_VCF',
		help = "Optional given vcf file. Enable to perform force calling. [NULL]",
		default = None,
        type = str)

	# **************Advanced Parameters******************
	GroupAdvanced = parser.add_argument_group('Advanced')

	# ++++++INS++++++
	GroupAdvanced.add_argument('--max_cluster_bias_INS', 
		help = "Maximum distance to cluster read together for insertion.[%(default)s]", 
		default = 100, 
		type = int)
	GroupAdvanced.add_argument('--diff_ratio_merging_INS', 
		help = "Do not merge breakpoints with basepair identity more than [%(default)s] for insertion.", 
		default = 0.3, 
		type = float)
	# GroupAdvanced.add_argument('--diff_ratio_filtering_INS', 
	# 	help = "Filter breakpoints with basepair identity less than [%(default)s] for insertion.", 
	# 	default = 0.6, 
	# 	type = float)

	# ++++++DEL++++++
	GroupAdvanced.add_argument('--max_cluster_bias_DEL', 
		help = "Maximum distance to cluster read together for deletion.[%(default)s]", 
		default = 200, 
		type = int)
	GroupAdvanced.add_argument('--diff_ratio_merging_DEL', 
		help = "Do not merge breakpoints with basepair identity more than [%(default)s] for deletion.", 
		default = 0.5, 
		type = float)
	# GroupAdvanced.add_argument('--diff_ratio_filtering_DEL', 
	# 	help = "Filter breakpoints with basepair identity less than [%(default)s] for deletion.", 
	# 	default = 0.7, 
	# 	type = float)

	# ++++++INV++++++
	GroupAdvanced.add_argument('--max_cluster_bias_INV', 
		help = "Maximum distance to cluster read together for inversion.[%(default)s]", 
		default = 500, 
		type = int)

	# ++++++DUP++++++
	GroupAdvanced.add_argument('--max_cluster_bias_DUP', 
		help = "Maximum distance to cluster read together for duplication.[%(default)s]", 
		default = 500, 
		type = int)

	# ++++++TRA++++++
	GroupAdvanced.add_argument('--max_cluster_bias_TRA', 
		help = "Maximum distance to cluster read together for translocation.[%(default)s]", 
		default = 50, 
		type = int)
	GroupAdvanced.add_argument('--diff_ratio_filtering_TRA', 
		help = "Filter breakpoints with basepair identity less than [%(default)s] for translocation.", 
		default = 0.6, 
		type = float)

	GroupAdvanced.add_argument('--remain_reads_ratio', 
		help = "The ratio of reads remained in cluster. Set lower when the alignment data have high quality but recommand over 0.5.[%(default)s]", 
		default = 1.0, 
		type = float)

	# parser.add_argument('-d', '--max_distance', 
	# 	help = "Maximum distance to group SV together..[%(default)s]", 
	# 	default = 1000, type = int)



	# These parameters are drawn lessons from pbsv v2.2.0
	# parser.add_argument('--min_del_size', 
	# 	help = "Minimum size of a deletion.[%(default)s]", 
	# 	default = 20, type = int)

	args = parser.parse_args(argv)
	return args

def Generation_VCF_header(file, contiginfo, sample, argv):
	# General header
	file.write("##fileformat=VCFv4.2\n")
	file.write("##source=cuteSV-%s\n"%(VERSION))
	import time
	file.write("##fileDate=%s\n"%(time.strftime('%Y-%m-%d %H:%M:%S %w-%Z',time.localtime())))
	for i in contiginfo:
		file.write("##contig=<ID=%s,length=%d>\n"%(i[0], i[1]))

	# Specific header
	# ALT
	file.write("##ALT=<ID=INS,Description=\"Insertion of novel sequence relative to the reference\">\n")
	file.write("##ALT=<ID=DEL,Description=\"Deletion relative to the reference\">\n")
	file.write("##ALT=<ID=DUP,Description=\"Region of elevated copy number relative to the reference\">\n")
	file.write("##ALT=<ID=INV,Description=\"Inversion of reference sequence\">\n")
	file.write("##ALT=<ID=BND,Description=\"Breakend of translocation\">\n")

	# INFO
	file.write("##INFO=<ID=PRECISE,Number=0,Type=Flag,Description=\"Precise structural variant\">\n")
	file.write("##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description=\"Imprecise structural variant\">\n")
	file.write("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">\n")
	file.write("##INFO=<ID=SVLEN,Number=1,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">\n")
	file.write("##INFO=<ID=CHR2,Number=1,Type=String,Description=\"Chromosome for END coordinate in case of a translocation\">\n")
	file.write("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">\n")
	file.write("##INFO=<ID=CIPOS,Number=2,Type=Integer,Description=\"Confidence interval around POS for imprecise variants\">\n")
	file.write("##INFO=<ID=CILEN,Number=2,Type=Integer,Description=\"Confidence interval around inserted/deleted material between breakends\">\n")
	# file.write("##INFO=<ID=MATEID,Number=.,Type=String,Description=\"ID of mate breakends\">\n")
	file.write("##INFO=<ID=RE,Number=1,Type=Integer,Description=\"Number of read support this record\">\n")
	file.write("##INFO=<ID=STRAND,Number=A,Type=String,Description=\"Strand orientation of the adjacency in BEDPE format (DEL:+-, DUP:-+, INV:++/--)\">\n")
	file.write("##INFO=<ID=RNAMES,Number=.,Type=String,Description=\"Supporting read names of SVs (comma separated)\">\n")
	file.write("##INFO=<ID=AF,Number=A,Type=Float,Description=\"Allele Frequency.\">\n")
	file.write("##FILTER=<ID=q5,Description=\"Quality below 5\">\n")
	# FORMAT
	# file.write("\n")
	file.write("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n")
	file.write("##FORMAT=<ID=DR,Number=1,Type=Integer,Description=\"# High-quality reference reads\">\n")
	file.write("##FORMAT=<ID=DV,Number=1,Type=Integer,Description=\"# High-quality variant reads\">\n")
	file.write("##FORMAT=<ID=PL,Number=G,Type=Integer,Description=\"# Phred-scaled genotype likelihoods rounded to the closest integer\">\n")
	file.write("##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"# Genotype quality\">\n")

	file.write("##CommandLine=\"cuteSV %s\"\n"%(" ".join(argv)))



# coding=utf-8

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name = "cuteSV",
    version = "2.1.0",
    description = "Long-read-based human genomic structural variation detection with cuteSV",
    author = "Jiang Tao",
    author_email = "tjiang@hit.edu.cn",
    url = "https://github.com/tjiangHIT/cuteSV",
    license = "MIT",
    packages = find_packages("src"),
    package_dir = {"": "src"},
    data_files = [("", ["LICENSE"])],
    scripts=['src/cuteSV/cuteSV'],
    # long_description = LONG_DESCRIPTION,
    long_description = readme,
    long_description_content_type = 'text/markdown',
    zip_safe = False,
    install_requires = ['scipy', 'Cython', 'pysam', 'Biopython', 'Cigar', 'numpy', 'pyvcf', 'scikit-learn']
)
